import sys
import traceback
from .bridge_globals import bridge_globals

def serialize(obj):
	return bridge_globals()['msg_service'].serializer.serialize(obj)

def deserialize(text):
	return bridge_globals()['msg_service'].serializer.deserialize(text)

def observer(commandId, observerId):
	return lambda obj: notify_observer(obj, commandId, observerId)

#### NOTIFICATION FUNCTIONS
def notify(obj, notificationId):
	bridge_globals()['logger'].log("PYTHON: Notify " + str(notificationId))
	data = {}
	data["type"] = "EVAL"
	data["id"] = notificationId
	data["value"] = serialize(obj)
	bridge_globals()['msg_service'].send_async_message(data)

def notify_observer(obj, commandId, observerId):
	bridge_globals()['logger'].log("PYTHON: Notify observer " + str(commandId) + " " + str(observerId))
	data = {}
	data["type"] = "CALLBACK"
	data["commandId"] = commandId
	data["observerId"] = observerId
	data["value"] = serialize(obj)
	rawValue = bridge_globals()['msg_service'].send_sync_message(data)['value']
	return deserialize(rawValue)

def notify_error(ex, command):
	bridge_globals()['logger'].log("Error on command: " + str(command.command_id()))
	bridge_globals()['logger'].log("Exception: " + str(ex))
	data = {}
	data["type"] = "ERR"
	data["errMsg"] = str(ex)
	data["trace"] = traceback.format_exc(100)
	data["commandId"] = command.command_id()
	return bridge_globals()['msg_service'].send_sync_message(data)

def bridge_inspect(obj):
	if hasattr(obj,'__dict__'):
		return obj.__dict__
	else:
		return {}



import json
import io
from .object_registry import registry

mapper = {}

def addMapping(key_type, mapping_function):
    mapper[key_type] = mapping_function

class JsonEncoder(json.JSONEncoder):
    def __init__(self, *args, **kwargs):
        json.JSONEncoder.__init__(self, *args, **kwargs)
        self.mapper = mapper

    def default(self, obj):
        if type(obj) in self.mapper:
            return self.mapper[type(obj)](obj)
        return registry().proxy(obj)

class JsonSerializer:
    def serialize(self, obj):
        return json.dumps(obj, cls=JsonEncoder)

    def deserialize(self, text):
        return json.loads(text)



from ...PythonBridge.object_registry import registry

class GtPhlowItemValue:
	def __init__(self, text = ""):
		self.itemText = text
		self.background = None

	def get_background(self):
		return self.background

	def set_background(self, colorDict):
		self.background = colorDict
		return self

	def get_itemText(self):
		return str(self.itemText)

	def set_itemText(self, text):
		self.itemText = text
		return self

	def asDictionaryForExport(self):
		data = {"itemText": self.get_itemText(), "valueTypeName": self.valueTypeName()}
		if (self.get_background() != None):
			data["background"] = self.get_background()
		return data

	def valueTypeName(self):
		return "item"

class GtPhlowItemErrorValue(GtPhlowItemValue):
	def background(self):
		return {"r": 1.0, "g": 0.4701857282502444, "b": 0.458455522971652}

	def valueTypeName(self):
		return "errorValue"

class GtPhlowItemTextualValue(GtPhlowItemValue):
	def valueTypeName(self):
		return "textualValue"

class GtPhlowRowValue:
	def __init__(self, values):
		self.columnValues = values

	def asDictionaryForExport(self):
		return {"columnValues": list(map(lambda each: each.asDictionaryForExport(), self.columnValues))}

class GtPhlowListingDataSource:
	def __init__(self, computation, accessor):
		self.computation = computation
		self.values = None
		self.accessor = accessor

	def rowValue(self, i):
		raise Exception("Not Implemented", "Subclasses should implement rowValue")

	def retrieveItems(self, count, startIndex):
		if (self.values == None):
			self.values = list(self.computation())
		result = []
		for i in range(startIndex - 1, min(len(self.values), startIndex + count)):
			result.append({"nodeValue": self.rowValue(i).asDictionaryForExport(), "nodeId": i})
		return result

	def retriveSentItemAt(self, index):
		return registry().proxy(self.values[index - 1] if self.accessor == None else self.accessor(index - 1))

	def flushItemsIterator(self):
		self.values = None

class GtPhlowListDataSource(GtPhlowListingDataSource):
	def __init__(self, computation, formatCallback, accessor):
		super().__init__(computation, accessor)
		self.formatCallback = formatCallback

	def rowValue(self, i):
		try:
			return GtPhlowItemTextualValue(self.formatCallback(self.values[i]))
		except Exception as ex:
			return GtPhlowItemErrorValue(str(ex))


class GtPhlowColumnedListDataSource(GtPhlowListingDataSource):
	def __init__(self, computation, columns, accessor):
		super().__init__(computation, accessor)
		self.columns = columns

	def getCellValue(self, row, column):
		try:
			return GtPhlowItemTextualValue(column.formatItem(row))
		except Exception as ex:
			return GtPhlowItemErrorValue(str(ex))

	def rowValue(self, i):
		try:
			row = self.values[i]
		except Exception as ex:
			return GtPhlowRowValue(list(map(lambda each: GtPhlowItemErrorValue(str(ex)), self.columns)))
		return self.rowValueOfRow(row)

	def rowValueOfRow(self, row):
		return GtPhlowRowValue(list(map(lambda each : self.getCellValue(row, each), self.columns)))

class GtPhlowColumnedTreeDataSource(GtPhlowColumnedListDataSource):
	def __init__(self, computation, columns, children, accessor):
		super().__init__(computation, columns, accessor)
		self.children = children

	def nodeForPath(self, path):
		current = self.values[path[0]]
		index = 1
		while (current != None and index < len(path)):
			current = list(self.children(current))[path[index]]
			index+=1
		return current

	def retriveSentItemAtPath(self, path):
		node = self.nodeForPath(path)
		return registry().proxy(node if self.accessor == None else self.accessor(node))

	def retrieveChildrenForNodeAtPath(self, path):
		result = []
		current = self.nodeForPath(path)
		nodes = list(self.children(current))
		for i in range(0, len(nodes)):
			result.append({"nodeValue": self.rowValueOfRow(nodes[i]).asDictionaryForExport(), "nodeId": i})
		return result



import threading

class StoppableThread(threading.Thread):   
    def __init__(self, loop_func, setup_func): 
        threading.Thread.__init__(self) 
        self._stop_event = threading.Event()
        self.daemon = True
        self.loop_func = loop_func
        self.setup_func = setup_func
  
    # function using _stop function 
    def stop(self): 
        self._stop_event.set() 
  
    def stopped(self): 
        return self._stop_event.isSet() 
  
    def run(self): 
        self.setup_func()
        while True: 
            if self.stopped(): 
                return
            self.loop_func()




from .view import View

class TextEditorView(View):
	def __init__(self):
		super().__init__()
		self.string = ""

	def setString(self, aString):
		self.string = aString
		return self
	
	def getString(self):
		return self.string
	
	def asDictionaryForExport(self):
		exportData = super().asDictionaryForExport()
		exportData["viewName"] = "GtPhlowTextEditorViewSpecification"
		exportData["dataTransport"] = 1
		exportData["string"] = self.getString()
		return exportData


from .view_columnedlist import ColumnedListView
from .data_source import GtPhlowColumnedTreeDataSource
from .column import ListColumn


class ColumnedTreeView(ColumnedListView):
	def __init__(self):
		super().__init__()
		self.childrenComputation = lambda each: []

	def children(self, children):
		self.childrenComputation = children
		return self

	def dataSource(self):
		return GtPhlowColumnedTreeDataSource(self.itemsCallback, self.columns, self.childrenComputation, self.accessor)
	
	def asDictionaryForExport(self):
		exportData = super().asDictionaryForExport()
		exportData["viewName"] = "GtPhlowColumnedTreeViewSpecification"
		return exportData
	